"""
JobResultSummary data class
"""

from dataclasses import dataclass, asdict
from typing import Union
from datetime import datetime

from .PMAsset import PMAssetJobView
from ...base.base_list import BaseList


@dataclass
class JobResultSummary:
    """
    Represents a job result summary in Patch Management.
    """

    id: str = None
    name: str = None
    assetCount: int = None
    patchCount: int = None
    createdBy: str = None
    createdOn: Union[int, datetime] = None
    assets: BaseList[PMAssetJobView] = None

    def __post_init__(self):
        if self.createdOn:
            setattr(self, "createdOn", datetime.fromtimestamp(self.createdOn / 1000))

        if self.assets:
            bl = BaseList()
            for asset in self.assets:
                asset_dict = dict()
                for key, value in asset.items():
                    if key == "asset":
                        asset_dict.update(value)
                    else:
                        asset_dict[key] = value

                asset_dict["jobId"] = self.id
                bl.append(PMAssetJobView.from_dict(**asset_dict))
            setattr(self, "assets", bl)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return asdict(self)

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    @classmethod
    def from_dict(cls, **kwargs):
        return cls(**kwargs)



"""
Patch Management call schema
"""

from frozendict import frozendict

PM_SCHEMA = frozendict(
    {
        "pm": {
            "url_type": "gateway",
            "deploymentjobs": {
                "endpoint": "/pm/v1/deploymentjobs/{placeholder}",
                "method": ["GET", "POST", "PATCH", "DELETE"],
                "valid_params": [
                    "placeholder",
                    "filter",
                    "attributes",
                    "platform",
                    "coauthorJob",
                    "ownedJob",
                    "pageNumber",
                    "pageSize",
                    "sort",
                ],
                "valid_POST_data": ["placeholder"],
                "use_requests_json_data": True,
                "return_type": "json",
                "pagination": True,
                "auth_type": "token",
            },
            "deploymentjob": {
                "endpoint": "/pm/v1/deploymentjob/{placeholder}",
                "method": ["GET", "POST", "PATCH", "DELETE"],
                "valid_params": ["placeholder"],
                "valid_POST_data": [
                    "placeholder",
                    "deploymentJobId",
                    "jobInstanceId",
                    "pageSize",
                    "sort",
                ],
                "use_requests_json_data": True,
                "return_type": "json",
                "pagination": False,
                "auth_type": "token",
            },
        },
    }
)



"""
Contains the data classes used for Patch Management Jobs.
"""

from dataclasses import dataclass, asdict
from typing import Union
from datetime import datetime
from string import digits

from .Tag import Tag
from ...base.base_list import BaseList


@dataclass
class PMJob:
    """
    Represents a deployment job in Patch Management.
    """

    id: str = None
    name: str = None
    coAuthorUserIds: BaseList[str] = None
    subCategory: str = None
    dynamicQQLType: int = None
    matchAllTags: BaseList[str] = None
    type: str = None
    recurringWeekDays: str = None
    totalAssetCount: int = None
    isDynamicPatchesQQL: bool = None
    assetCount: Union[int, None] = None
    customPatchUrlConfigured: bool = None
    linkedJobReferenceCount: Union[int, None] = None
    created: dict = None
    # created is parsed into below fields:
    created_date: Union[int, datetime] = None
    created_user_name: str = None
    created_user_id: str = None
    # end created
    triggerStatus: int = None
    recurringDayOfMonth: str = None
    jobSource: int = None
    readOnly: bool = None
    tags: BaseList[Tag] = None
    exclusionAssetIds: BaseList[str] = None
    scheduleType: str = None
    dayOfMonth: str = None
    disabledPatches: BaseList[str] = None
    filterType: str = None
    updated: dict = None
    # updated is parsed into below fields:
    updated_date: Union[int, datetime] = None
    updated_user_name: str = None
    updated_user_id: str = None
    # end updated
    additionalQQLS: BaseList[str] = None
    status: str = None
    timezone: str = None
    exclusionFilterType: str = None
    nextScheduleDateTime: Union[str, datetime] = None
    platform: str = None
    mitigationActionCount: int = None
    isPriorityJob: bool = None
    customerId: str = None
    exclusionTags: BaseList[Tag] = None
    taggedAssetCount: int = None
    patchCount: int = None
    applicableAssetCount: int = None
    patchTuesdayPlusXDays: int = None
    schemaVersion: float = None
    recurringLastDayOfMonth: bool = None
    recurringWeekDayOfMonth: str = None
    isRecurring: bool = None
    lastScheduleDateTime: Union[str, datetime] = None
    dynamicPatchesQQL: str = None
    startDateTime: Union[str, datetime] = None
    timezoneType: str = None
    deleted: bool = None
    monthlyRecurringType: int = None
    completionPercent: float = None
    category: int = None
    linkedJobId: str = None
    assetResultReceivedCount: int = None
    isVulnContext: bool = None
    isAssetImported: bool = None
    remediationQids: BaseList[str] = None

    def __post_init__(self):
        # Define attributes that need to be converted
        # in a specific way

        TO_INT_FIELDS = [
            "totalAssetCount",
            "assetResultReceivedCount",
            "applicableAssetCount",
        ]

        FROM_TIMESTAMP_FIELDS = [
            "nextScheduleDateTime",
            "lastScheduleDateTime",
        ]

        BREAKDOWN_DT_FIELDS = [
            "created",
            "updated",
        ]

        TAG_OBJ_FIELDS = [
            "tags",
            "exclusionTags",
        ]

        NON_STR_OR_OBJ_BASELIST_FIELDS = [
            "remediationQids",
        ]

        """
        CURRENT FIELDS THAT I HAVE NOT BEEN ABLE TO
        DETERMINE THE STRUCTURE OF:

        matchAllTags
        assetCount
        exclusionAssetIds
        disabledPatches
        additionalQQLS
        recurringLastDayOfMonth
        """

        if self.schemaVersion:
            setattr(self, "schemaVersion", float(self.schemaVersion))

        if self.completionPercent and all(
            [char in digits for char in str(self.completionPercent)]
        ):
            setattr(self, "completionPercent", float(self.completionPercent))

        if self.coAuthorUserIds:
            setattr(self, "coAuthorUserIds", BaseList(self.coAuthorUserIds))

        if self.startDateTime:
            setattr(
                self,
                "startDateTime",
                datetime.strptime(self.startDateTime, "%Y-%m-%d %I:%M:%S %p"),
            )

        for base_list_field in NON_STR_OR_OBJ_BASELIST_FIELDS:
            if getattr(self, base_list_field):
                setattr(self, base_list_field, BaseList(getattr(self, base_list_field)))

        for tag_obj in TAG_OBJ_FIELDS:
            if getattr(self, tag_obj):
                setattr(
                    self,
                    tag_obj,
                    BaseList([Tag.from_dict(tag) for tag in getattr(self, tag_obj)]),
                )

        for dt_field in BREAKDOWN_DT_FIELDS:
            if getattr(self, dt_field):
                setattr(
                    self,
                    f"{dt_field}_date",
                    datetime.fromtimestamp(getattr(self, dt_field)["date"] / 1000),
                )
                setattr(
                    self,
                    f"{dt_field}_user_name",
                    getattr(self, dt_field)["user"].get("name"),
                )
                setattr(
                    self,
                    f"{dt_field}_user_id",
                    getattr(self, dt_field)["user"].get("id"),
                )
                setattr(self, dt_field, None)

        for timestamp_field in FROM_TIMESTAMP_FIELDS:
            if getattr(self, timestamp_field):
                setattr(
                    self,
                    timestamp_field,
                    datetime.fromtimestamp(getattr(self, timestamp_field) / 1000),
                )

        for int_field in TO_INT_FIELDS:
            if getattr(self, int_field):
                setattr(self, int_field, int(getattr(self, int_field)))

    @classmethod
    def from_dict(cls, **kwargs):
        return cls(**kwargs)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()



"""
Cloud Agent API call schema
"""

from frozendict import frozendict

CLOUDAGENT_SCHEMA = frozendict(
    {
        "cloud_agent": {
            "url_type": "api",
            "purge_agent": {
                "endpoint": "/qps/rest/2.0/uninstall/am/asset/{placeholder}",  # assetId
                "method": ["POST"],
                "valid_params": [],
                "valid_POST_data": ["placeholder", "_xml_data"],
                "use_requests_json_data": False,
                "return_type": "xml",
                "pagination": False,
                "auth_type": "basic",
                "_xml_data": True,
            },
            "bulk_purge_agent": {
                "endpoint": "/qps/rest/2.0/uninstall/am/asset",
                "method": ["POST"],
                "valid_params": [],
                "valid_POST_data": ["_xml_data"],
                "use_requests_json_data": False,
                "return_type": "xml",
                "pagination": False,
                "auth_type": "basic",
                "_xml_data": True,
            },
            "list_agents": {
                "endpoint": "/qps/rest/2.0/search/am/hostasset",
                "method": ["POST"],
                "valid_params": [],
                "valid_POST_data": ["_xml_data"],
                "use_requests_json_data": False,
                "return_type": "xml",
                "pagination": True,
                "auth_type": "basic",
                "_xml_data": True,
            },
            "launch_ods": {
                "endpoint": "/qps/rest/1.0/ods/ca/agentasset/{placeholder}",
                "method": ["POST"],
                "valid_params": ["scan", "overrideConfigCpu"],
                "valid_POST_data": ["placeholder", "_xml_data"],
                "use_requests_json_data": False,
                "return_type": "xml",
                "pagination": False,
                "auth_type": "basic",
                "_xml_data": True,
            },
            "bulk_launch_ods": {
                "endpoint": "/qps/rest/1.0/ods/ca/agentasset",
                "method": ["POST"],
                "valid_params": ["scan", "overrideConfigCpu"],
                "valid_POST_data": ["_xml_data"],
                "use_requests_json_data": False,
                "return_type": "xml",
                "pagination": False,
                "auth_type": "basic",
                "_xml_data": True,
            },
        },
    }
)



"""
call_schema.py - contains the CALL_SCHEMA lookup for the qualysdk package.

The CALL_SCHEMA dictionary contains the schema for each Qualys API call. 
This schema is used to determine what parameters are required for each call and where/how
they should be sent to the API.

To view each module's schema, look in ./call_schemas/
"""

from frozendict import frozendict

from .call_schemas import *

schemas = [
    CLOUDAGENT_SCHEMA,
    CS_SCHEMA,
    GAV_SCHEMA,
    PM_SCHEMA,
    TOTALCLOUD_SCHEMA,
    VMDR_SCHEMA,
    WAS_SCHEMA,
]

CALL_SCHEMA = dict()

for s in schemas:
    CALL_SCHEMA.update(s)

CALL_SCHEMA = frozendict(CALL_SCHEMA)



"""
Container security call schemas
"""

from frozendict import frozendict

CS_SCHEMA = frozendict(
    {
        "containersecurity": {
            "url_type": "gateway",
            "list_containers": {
                "endpoint": "/csapi/v1.3/containers/list",
                "method": ["GET"],
                "valid_params": ["filter", "paginationQuery", "limit"],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": True,
                "auth_type": "token",
            },
            "get_container_details": {
                "endpoint": "/csapi/v1.3/containers/{placeholder}",
                "method": ["GET"],
                "valid_params": ["placeholder", "containerSha"],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "token",
            },
        },
    }
)



"""
Represents a Qualys Asset in Patch Management.
"""

from dataclasses import dataclass, asdict
from typing import Union
from datetime import datetime

from .Tag import Tag
from ...base.base_list import BaseList


@dataclass
class PMInterface:
    """
    Network Interface Data
    """

    interfaceName: str = None
    macAddress: str = None
    address: str = None
    gatewayAddress: str = None

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return asdict(self)

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    def __str__(self):
        return self.address

    @classmethod
    def from_dict(cls, **kwargs):
        return cls(**kwargs)


@dataclass
class PMAssetJobView:
    """
    Represents a Qualys Asset in Patch Management
    with additional job information attached.
    """

    id: str = None
    name: str = None
    operatingSystem: str = None
    jobId: str = None
    tags: BaseList[Tag] = None
    lastLoggedOnUser: str = None
    successPatches: int = None
    installedPatches: int = None
    failedPatches: int = None
    supersededPatches: int = None
    notApplicablePatches: int = None
    executing: Union[int, bool] = None
    pendingExecution: Union[int, bool] = None
    pendingReboot: Union[int, bool] = None
    pendingVerification: Union[int, bool] = None
    jobInstanceId: str = None
    interfaces: BaseList[PMInterface] = None
    skipPatchCount: int = None
    additionalFields: dict = None
    endDateTime: Union[int, datetime] = None
    startDateTime: Union[int, datetime] = None
    statusDateTime: Union[int, datetime] = None
    status: str = None
    statusCode: str = None
    jobSentOn: Union[int, datetime] = None
    installed: BaseList[str] = None
    failed: BaseList[str] = None
    success: BaseList[str] = None
    superseded: BaseList[str] = None
    notApplicable: BaseList[str] = None
    failedActionsCount: int = None
    successfulActionsCount: int = None
    skippedActionsCount: int = None
    interimResultStatus: str = None
    totalPatchCount: int = None
    runId: int = None
    scanDateTime: Union[int, datetime] = None
    pendingForRebootInAnotherJob: bool = None
    pendingForRebootInAnotherJobName: str = None
    osIdentifier: str = None

    def __post_init__(self):
        TO_BL_FIELDS = [
            "installed",
            "failed",
            "success",
            "superseded",
            "notApplicable",
        ]

        for bl_field in TO_BL_FIELDS:
            if getattr(self, bl_field):
                bl = BaseList()
                for obj in getattr(self, bl_field):
                    bl.append(obj)
                setattr(self, bl_field, bl)

        FROM_TIMESTAMP_FIELDS = [
            "endDateTime",
            "startDateTime",
            "statusDateTime",
            "jobSentOn",
            "scanDateTime",
        ]

        for timestamp_field in FROM_TIMESTAMP_FIELDS:
            if getattr(self, timestamp_field):
                setattr(
                    self,
                    timestamp_field,
                    datetime.fromtimestamp(getattr(self, timestamp_field) / 1000),
                )

        TO_BOOL_FIELDS = [
            "executing",
            "pendingExecution",
            "pendingReboot",
            "pendingVerification",
        ]

        for bool_field in TO_BOOL_FIELDS:
            if getattr(self, bool_field):
                setattr(self, bool_field, bool(getattr(self, bool_field)))

        if self.interfaces:
            bl = BaseList()
            for interface in self.interfaces:
                bl.append(PMInterface.from_dict(**interface))
            setattr(self, "interfaces", bl)

        if self.tags:
            bl = BaseList()
            for tag in self.tags:
                bl.append(Tag.from_dict(tag))
            setattr(self, "tags", bl)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return asdict(self)

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    def __str__(self):
        return self.name

    @classmethod
    def from_dict(cls, **kwargs):
        return cls(**kwargs)



"""
CLI script to quickly perform Patch Management
(PM) operations using qualysdk.
"""

from argparse import ArgumentParser, Namespace

from qualysdk import TokenAuth, write_excel, BaseList, write_json
from qualysdk.pm import *


def cli_fn(auth: TokenAuth, args: Namespace, endpoint: str) -> None:
    kwargs = dict(args.kwarg) if args.kwarg else {}
    if "page_count" in kwargs:
        kwargs["page_count"] = int(kwargs["page_count"])
    for kwarg in kwargs:
        if str(kwargs[kwarg]).lower() == "true":
            kwargs[kwarg] = True
        elif str(kwargs[kwarg]).lower() == "false":
            kwargs[kwarg] = False

    match endpoint:
        case "list_jobs":
            result = list_jobs(auth, **kwargs)
        case "get_job_results":
            result = get_job_results(auth, args.job_id, **kwargs)
        case "get_job_runs":
            result = get_job_runs(auth, args.job_id, **kwargs)
        case _:
            raise ValueError(f"Invalid endpoint: {endpoint}.")

    # If the result object does NOT have the len() method available,
    # we need to wrap it in a BaseList:
    if not hasattr(result, "__len__"):
        bl = BaseList()
        bl.append(result)
        result = bl

    write_excel(result, args.output)


def main():
    parser = ArgumentParser(
        description="CLI script to quickly perform Patch Management (PM) operations using qualysdk"
    )
    parser.add_argument(
        "-u", "--username", required=True, help="Qualys username", type=str
    )
    parser.add_argument(
        "-p", "--password", required=True, help="Qualys password", type=str
    )
    parser.add_argument(
        "-P",
        "--platform",
        help="Qualys platform",
        default="qg3",
        choices=["qg1", "qg2", "qg3", "qg4"],
    )

    # subparser for action:
    subparsers = parser.add_subparsers(dest="action", help="Action to perform")

    # get_findings action:
    list_jobs_parser = subparsers.add_parser("list_jobs", help="Get a list of PM jobs.")
    list_jobs_parser.add_argument(
        "-o",
        "--output",
        help="Output xlsx file to write results to",
        type=str,
        default="pm_jobs.xlsx",
    )
    list_jobs_parser.add_argument(
        "--kwarg",
        help="Specify a keyword argument to pass to the action. Can be used multiple times",
        action="append",
        nargs=2,
        metavar=("key", "value"),
    )

    job_results_parser = subparsers.add_parser(
        "get_job_results", help="Get results for a PM job."
    )
    job_results_parser.add_argument(
        "-o",
        "--output",
        help="Output xlsx file to write results to",
        type=str,
        default="pm_job_{JOB_ID}_results.xlsx",
    )
    job_results_parser.add_argument(
        "-j",
        "--job-id",
        help="Specify the job ID to get results for. Can be used multiple times",
        type=str,
        required=True,
        action="append",
    )
    job_results_parser.add_argument(
        "--kwarg",
        help="Specify a keyword argument to pass to the action. Can be used multiple times",
        action="append",
        nargs=2,
        metavar=("key", "value"),
    )

    job_runs_parser = subparsers.add_parser(
        "get_job_runs", help="Get runs for a PM job."
    )
    job_runs_parser.add_argument(
        "-o",
        "--output",
        help="Output xlsx file to write results to",
        type=str,
        default="pm_job_{JOB_ID}_runs.xlsx",
    )
    job_runs_parser.add_argument(
        "-j",
        "--job-id",
        help="Specify the job ID to get runs for",
        type=str,
        required=True,
    )
    job_runs_parser.add_argument(
        "--kwarg",
        help="Specify a keyword argument to pass to the action. Can be used multiple times",
        action="append",
        nargs=2,
        metavar=("key", "value"),
    )

    args = parser.parse_args()

    # create TokenAuth object
    auth = TokenAuth(args.username, args.password, platform=args.platform)

    match args.action:
        case "list_jobs":
            cli_fn(auth=auth, args=args, endpoint="list_jobs")
        case "get_job_results":
            if len(args.job_id) > 1:
                args.output = args.output.replace("{JOB_ID}", "multiple_jobs")
            else:
                args.output = args.output.replace("{JOB_ID}", args.job_id[0])
            cli_fn(auth=auth, args=args, endpoint="get_job_results")
        case "get_job_runs":
            args.output = args.output.replace("{JOB_ID}", args.job_id)
            cli_fn(auth=auth, args=args, endpoint="get_job_runs")
        case _:
            parser.print_help()
            exit(1)


if __name__ == "__main__":
    main()



"""
Contains resource dataclasses for Azure, such as SQL, VMs, etc.
"""

from dataclasses import dataclass, asdict
from typing import Union
from datetime import datetime

from ...base.base_list import BaseList
from .QID import QID


def process_file_or_blob(self, data, prefix):
    dict_keys = ["lastEnabledTime", "keyType", "enabled"]
    for field in dict_keys:
        if data.get(field):
            setattr(self, f"{prefix}_{field}", data.get(field))


@dataclass
class BaseResource:
    """
    Base dataclass for all AWS resources in TotalCloud
    """

    resourceId: str = None
    name: str = None
    connectorUuids: BaseList[str] = None
    created: Union[str, datetime] = None
    subscriptionName: str = None
    subscriptionId: str = None
    uuid: str = None
    connectorUuid: str = None
    tags: BaseList[str] = None
    remediationEnabled: bool = None
    updated: Union[str, datetime] = None
    additionalDetails: dict = None
    cloudType: str = None
    region: str = None
    resourceType: str = None
    controlsFailed: int = None
    qualysTags: BaseList[str] = None
    customerId: str = None
    customers: BaseList[str] = None
    resourceGroupName: str = None
    scanUuid: str = None
    _type: str = None

    def __post_init__(self):
        """
        __post_init__ - post initialization method for the BaseResource dataclass
        """
        DT_FIELDS = ["created", "updated"]
        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        if self.additionalDetails:
            setattr(self, "additionalDetails", str(self.additionalDetails))

        if self.connectorUuids:
            data = self.connectorUuids
            bl = BaseList()
            for uuid in data:
                bl.append(uuid)
            setattr(self, "connectorUuids", bl)

        if self.tags:
            data = self.tags
            bl = BaseList()
            if isinstance(data, dict):
                data = [data]
            for tag in data:
                bl.append(f"{tag.get('value', None)}")
            setattr(self, "tags", bl)

        if self.qualysTags:
            data = self.qualysTags
            bl = BaseList()
            if isinstance(data, dict):
                data = [data]
            for tag in data:
                bl.append(tag["tagName"])
            setattr(self, "qualysTags", bl)

        if self.remediationEnabled:
            data = self.remediationEnabled
            if not isinstance(data, bool):
                setattr(self, "remediationEnabled", data.lower() == "true")

    def to_dict(self):
        """
        to_dict - return the BaseResource as a dictionary
        """
        return asdict(self)

    def keys(self):
        """
        keys - return the keys of the BaseResource
        """
        return self.to_dict().keys()

    def values(self):
        """
        values - return the values of the BaseResource
        """
        return self.to_dict().values()

    def items(self):
        """
        items - return the items of the BaseResource
        """
        return self.to_dict().items()

    def __dict__(self):
        """
        __dict__ - return the BaseResource as a dictionary
        """
        return self.to_dict()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create a BaseResource object from a dictionary
        """
        return BaseResource(**data)


@dataclass
class AzureVM(BaseResource):
    """
    Represents an Azure VM in TotalCloud
    """

    arsScore: int = None
    vulnerabilityStats: dict = None
    computerName: str = None
    primaryPrivateIPAddress: str = None
    primaryPublicIPAddress: str = None
    primaryPublicIPAddressId: str = None
    statuses: BaseList[str] = None
    criticalityScore: int = None
    availabilitySetId: str = None
    provisioningState: str = None
    licenseType: str = None
    size: str = None
    osType: str = None
    networkSecurityGroupId: str = None
    imageData: BaseList[str] = None
    resourceType: str = "VIRTUAL_MACHINE"

    def __post_init__(self):
        """
        __post_init__ - post initialization method for the AzureVM dataclass
        """
        super().__post_init__()

        if self.vulnerabilityStats:
            data = self.vulnerabilityStats
            setattr(self, "vulnerabilityStats", QID(**data))

        if self.statuses:
            data = self.statuses
            bl = BaseList()
            for status in data:
                bl.append(
                    f"{status.get('code')}:{status.get('level')}:{status.get('displayStatus')}:{status.get('message')}:{status.get('time')}"
                )
            setattr(self, "statuses", bl)

        if self.imageData:
            data = self.imageData
            bl = BaseList()
            for image in data:
                bl.append(
                    f"{image.get('id')}:{image.get('offer')}:{image.get('publisher')}:{image.get('version')} (sku:{image.get('sku')})"
                )
            setattr(self, "imageData", bl)

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AzureVM object from a dictionary
        """
        return AzureVM(**data)


@dataclass
class AzureWebApp(BaseResource):
    """
    Represents an Azure Web App in TotalCloud
    """

    redundancyMode: str = None
    repositorySiteName: str = None
    enabled: bool = None
    enabledHosts: BaseList[str] = None
    clientCertEnabled: bool = None
    subKinds: BaseList[str] = None
    usageState: str = None
    isDefaultContainer: bool = None
    appServicePlan: dict = None
    deploymentId: str = None
    httpsonly: bool = None
    clientAffinityEnabled: bool = None
    state: str = None
    key: str = None
    defaultHostName: str = None
    availabilityState: str = None
    kind: str = None
    resourceType: str = "WEB_APP"

    def __post_init__(self):
        super().__post_init__()

        if self.enabledHosts:
            data = self.enabledHosts
            bl = BaseList()
            for host in data:
                bl.append(host)
            setattr(self, "enabledHosts", bl)

        if self.subKinds:
            data = self.subKinds
            bl = BaseList()
            for kind in data:
                bl.append(kind)
            setattr(self, "subKinds", bl)

    @staticmethod
    def from_dict(data):
        return AzureWebApp(**data)


@dataclass
class AzureStorageAccount(BaseResource):
    """
    Represents an Azure storage account in TotalCloud
    """

    displayName: str = None
    connectorId: str = None
    scanId: str = None
    skuTier: str = None
    resourceIdentity: dict = None
    # resourceIdentity is parsed into below fields:
    resourceIdentity_type: str = None
    # end of resourceIdentity fields
    minimumTlsVersion: str = None
    kind: str = None
    firstDiscoveredOn: Union[str, datetime] = None
    lastDiscoveredOn: Union[str, datetime] = None
    skuName: str = None
    file: dict = None
    # file is parsed into below fields:
    file_lastEnabledTime: Union[str, datetime] = None
    file_keyType: str = None
    file_enabled: bool = None
    # end of file fields
    blob: dict = None
    # blob is parsed into below fields:
    blob_lastEnabledTime: Union[str, datetime] = None
    blob_keyType: str = None
    blob_enabled: bool = None
    # end of blob fields
    primaryLocation: str = None
    secondaryLocation: str = None
    hnsEnabled: Union[str, bool] = None
    resourceGroupId: str = None
    supportsHttpsTrafficOnly: bool = None
    statusOfPrimary: str = None
    statusOfSecondary: str = None
    location: str = None
    networkAcls: dict = None
    # networkAcls is parsed into below fields:
    networkAcls_bypass: str = None
    networkAcls_defaultAction: str = None
    networkAcls_ipRules: BaseList[str] = None
    networkAcls_virtualNetworkRules: BaseList[str] = None
    # end of networkAcls fields

    def __post_init__(self):
        super().__post_init__()

        DT_FIELDS = [
            "firstDiscoveredOn",
            "lastDiscoveredOn",
            "file_lastEnabledTime",
            "blob_lastEnabledTime",
        ]

        if self.resourceIdentity:
            dict_keys = ["type"]
            for field in dict_keys:
                if self.resourceIdentity.get(field):
                    setattr(
                        self,
                        f"resourceIdentity_{field}",
                        self.resourceIdentity.get(field),
                    )
            setattr(self, "resourceIdentity", None)

        if self.file:
            process_file_or_blob(self, self.file, "file")
            setattr(self, "file", None)

        if self.blob:
            process_file_or_blob(self, self.blob, "blob")
            setattr(self, "blob", None)

        if self.hnsEnabled:
            if isinstance(self.hnsEnabled, str):
                setattr(self, "hnsEnabled", self.hnsEnabled.lower() == "true")
            else:
                setattr(self, "hnsEnabled", self.hnsEnabled)

        if self.networkAcls:
            dict_keys = ["bypass", "defaultAction", "ipRules", "virtualNetworkRules"]
            ipRules_list = BaseList()
            virtualNetworkRules_list = BaseList()
            for field in dict_keys:
                if self.networkAcls.get(field):
                    if field == "ipRules":
                        for ipRule in self.networkAcls.get(field):
                            ipRules_list.append(
                                f"{ipRule.get('action')} {ipRule.get('value')}"
                            )
                    elif field == "virtualNetworkRules":
                        for virtualNetworkRule in self.networkAcls.get(field):
                            # TODO: parse virtualNetworkRule
                            virtualNetworkRules_list.append(virtualNetworkRule)
                    else:
                        setattr(
                            self, f"networkAcls_{field}", self.networkAcls.get(field)
                        )
            setattr(self, "networkAcls_ipRules", ipRules_list)
            setattr(self, "networkAcls_virtualNetworkRules", virtualNetworkRules_list)
            setattr(self, "networkAcls", None)

        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

    @staticmethod
    def from_dict(data):
        return AzureStorageAccount(**data)



"""
PMRun class shows high-level information about a
Patch Management run of a job
"""

from dataclasses import dataclass, asdict
from typing import Union
from datetime import datetime


@dataclass
class PMRun:
    """
    Represents a run of a Patch Management job.
    """

    jobInstanceId: int = None
    jobId: str = None
    scheduledDateTime: Union[str, datetime] = None
    timezoneType: str = None

    def __post_init__(self):
        if self.scheduledDateTime and not isinstance(self.scheduledDateTime, datetime):
            setattr(
                self,
                "scheduledDateTime",
                datetime.fromisoformat(self.scheduledDateTime),
            )

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return asdict(self)

    def __int__(self):
        return asdict(self)

    def keys(self):
        return asdict(self).keys()

    def values(self):
        return asdict(self).values()

    def items(self):
        return asdict(self).items()

    @classmethod
    def from_dict(cls, **kwargs):
        return cls(**kwargs)
